# coding: utf-8
from copy import copy

from .elements import BaseElement, AnonymusElement
from .exceptions import ElementNameConflict


class ElementRegistry:
    def __init__(self):
        self.xml_to_object_mapper = {}
        self.anonymus_element_klass = AnonymusElement

    def element(self, name=None, element_class=None, namespace=None):
        if namespace:
            fqdn = '{' + namespace + '}' + name
        else:
            fqdn = name

        if name is not None and fqdn in self.xml_to_object_mapper.keys():
            raise ElementNameConflict(element_class.__name__)

        if name is None and element_class is None:
            # @registry.tag()
            return self.element_function
        elif name is not None and element_class is None:
            if callable(name):
                # @registry.tag
                return self.element_function(name)
            else:
                # @registry.tag('somename') or @registry.tag(name='somename')
                def dec(func):
                    return self.element(fqdn, func, namespace=None)
                return dec
        elif name is not None and element_class is not None:
            self.xml_to_object_mapper[fqdn] = element_class
            return element_class

    def element_function(self, element_class):
        self.xml_to_object_mapper[element_class.__name__] = element_class
        return element_class

    def get_class_by_name(self, element_name: str) -> type:
        if element_name not in self.xml_to_object_mapper.keys():
            return BaseElement

        return self.xml_to_object_mapper[element_name]

    def __copy__(self):
        other = ElementRegistry()
        other.xml_to_object_mapper = copy(self.xml_to_object_mapper)

        return other


registry = ElementRegistry()



from unittest import TestCase

from django.template.backends.dummy import Template
from django.test import override_settings
from mock import Mock, PropertyMock, patch
from livesync.core.middleware import DjangoLiveSyncMiddleware


class LiveSyncMiddlewareTestCase(TestCase):
    def setUp(self):
        self.mocked_response = Mock()
        self.get_response = Mock(return_value=self.mocked_response)
        self.middleware = DjangoLiveSyncMiddleware(self.get_response)
        self.mock_content_property = PropertyMock(return_value=b"<body></body>")
        type(self.mocked_response).content = self.mock_content_property

    @override_settings(DEBUG=True)
    def test_middleware_injects_js_file_correctly(self):
        self.mocked_response.__getitem__ = Mock(return_value='text/html')
        # act
        self.middleware(Mock())
        # assert

        self.mock_content_property.assert_called()
        self.assertTrue(b"<script src='/static/livesync.js'></script>" in self.mock_content_property.call_args[0][0])

    @override_settings(DEBUG=True)
    def test_middleware_does_not_inject_js_file_if_content_type_is_not_html(self):
        self.mocked_response.__getitem__ = Mock(return_value='json')
        # act
        self.middleware(Mock())
        # assert
        self.mock_content_property.assert_not_called()

    @override_settings(DEBUG=False)
    def test_middleware_does_not_inject_js_file_if_not_debugging(self):
        self.mocked_response.__getitem__ = Mock(return_value='text/html')
        # act
        self.middleware(Mock())
        # assert
        self.mock_content_property.assert_not_called()

    @override_settings(DEBUG=True)
    @patch('django.template.loader.get_template')
    def test_middleware_custom_snippet_injektion(self, mock_template):
        self.mocked_response.__getitem__ = Mock(return_value='text/html')

        mock_template.return_value = Template('mocked_snippet')
        # act
        self.middleware(Mock())
        # assert

        self.mock_content_property.assert_called()
        self.assertTrue(b"mocked_snippet" in self.mock_content_property.call_args[0][0])



import datetime

from django import forms
from django.forms import widgets
from django_kirui.widgets import CheckboxSwitch, SimpleFileInput
from kirui_devserver.backend.models import Topic, Activity


class SampleForm(forms.ModelForm):
    elso = forms.CharField(label='Első mező', required=True, disabled=False, initial='charfield')
    masodik = forms.ChoiceField(label='Második mező', choices=[(None, '-----------------'), (1, 'One'), (2, 'Two'), (3, 'Three')], initial=2)
    harmadik = forms.MultipleChoiceField(label='Harmadik', choices=[(1, "One\\aaaasdfsdfsdf'"), (2, 'Two'), (3, 'Three')], initial=[2, 3],
                                         widget=widgets.CheckboxSelectMultiple, disabled=False, required=True)
    negyedik = forms.BooleanField(label='Negyedik mező', widget=CheckboxSwitch, disabled=False, initial=True)
    otodik = forms.CharField(label='Ötödik', widget=widgets.Textarea, initial='<b>asdf</b>', disabled=False)
    hatodik = forms.IntegerField(label='Hatodik', disabled=False)
    hetedik = forms.DateField(label='Hetedik', disabled=False, initial=datetime.date.today())
    nyolcadik = forms.FileField(label='Nyolcadik', widget=SimpleFileInput)
    hetedik_b = forms.DateField(label='Hetedik (b)', disabled=False, initial=datetime.date.today())

    class Meta:
        model = Activity
        fields = [] #'topics', 'reason'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        #for i in range(0, 80):
        #    self.fields[f'field_{i}'] = forms.BooleanField(label='Negyedik mező', disabled=False)  # widget=CheckboxSwitch,
        # self.fields['elso'].initial = str(datetime.datetime.now())


class FilterForm(forms.Form):
    field = forms.CharField(label='Filter field', required=False)



from django.conf import settings
from django.core.management.base import BaseCommand
from django.template.loader import get_template
from django.utils.module_loading import import_string

from kirui_devserver.backend.forms import SampleForm
from kirui_devserver.backend.models import Activity


class Command(BaseCommand):
    def handle(self, *args, **options):
        from django.test.client import RequestFactory
        rf = RequestFactory()
        post_request = rf.post('/backend/form/', {'foo': 'bar'})
        form = SampleForm(post_request, instance=Activity.objects.first())
        klass = getattr(settings, 'DJANGO_SAMON_BINDING_CLASS', 'django_kirui.context_processors.DjangoSamonBinding')
        klass = import_string(klass)

        data = get_template('xml/form.html').render(context={'form': form, 'djsamon': klass(post_request)}, request=None, to='psx')
        print(data)



from django.db import models


class Topic(models.Model):
    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


class Reason(models.Model):
    name = models.CharField(max_length=200)

    def __str__(self):
        return self.name


class Activity(models.Model):
    topics = models.ManyToManyField(Topic)
    reason = models.ForeignKey(Reason, on_delete=models.CASCADE, blank=True, null=True)



import unittest
from django.conf import settings
from django.core.management.base import CommandError
from django.test import override_settings

from livesync.management.commands.runserver import Command
from livesync.core.handler import LiveReloadRequestHandler


class MockEventHandler:
    @property
    def watched_paths(self):
        return []


class OptionsParserTestcase(unittest.TestCase):
    @override_settings(DJANGO_LIVESYNC={})
    def test_parse_default_options(self):
        command = Command()
        options = {'liveport': None, 'addrport': ''}
        command._parse_options(**options)

        self.assertEqual(command.liveport, 9001)
        self.assertEqual(command.livehost, 'localhost')

    def test_parse_custom_port_and_host(self):
        command = Command()
        options = {'liveport': 8888, 'addrport': '0.0.0.0:8000'}
        command._parse_options(**options)

        self.assertEqual(command.liveport, 8888)
        self.assertEqual(command.livehost, '0.0.0.0')

    def test_parse_invalid_port_raises_command_error(self):
        command = Command()
        options = {'liveport': 'abc', 'addrport': ''}

        with self.assertRaises(CommandError):
            command._parse_options(**options)

    def test_default_event_handler(self):
        command = Command()
        command._start_watchdog()
        handler, = command.file_watcher.handlers
        self.assertIsInstance(handler, LiveReloadRequestHandler)

    @override_settings(DJANGO_LIVESYNC={'EVENT_HANDLER': 'livesync.tests.test_command.MockEventHandler'})
    def test_customize_event_handler(self):
        command = Command()
        command._start_watchdog()
        handler, = command.file_watcher.handlers
        self.assertIsInstance(handler, MockEventHandler)



import ast
import json
from contextlib import contextmanager
from io import StringIO
from typing import TYPE_CHECKING

from . import constants
from .expressions import Bind

if TYPE_CHECKING:
    from .elements import BaseElement, AnonymusElement


class RenderedElement:
    node_attributes: dict

    def __init__(self, element: 'BaseElement', context: dict) -> None:
        self._element = element
        self._context = context
        self.node_attributes = self._eval_node_attributes(context)

    @property
    def node_name(self) -> str:
        return self._element.xml_tag

    def _eval_node_attributes(self, context) -> dict:
        retval = {}
        for k, v in self._element.xml_attrs.items():
            if k.startswith(f'{{{constants.XML_NAMESPACE_FLOW_CONTROL}}}') or k == constants.XML_TEXT_BINDING_TAG:
                continue
            elif k.startswith(f'{{{constants.XML_NAMESPACE_DATA_BINDING}}}') and hasattr(v, 'eval'):
                k = k.replace(f'{{{constants.XML_NAMESPACE_DATA_BINDING}}}', '')

                retval[k] = v.eval(context)
            elif k.startswith(f'{{{constants.XML_NAMESPACE_POJO}}}'):
                retval[k] = v.eval(context)
            else:
                retval[k] = v

        return retval

    @property
    def children(self):
        if text := self._element.xml_attrs.get(f'{constants.XML_TEXT_BINDING_TAG}', None):
            yield text.eval(context=self._context)
            return

        for child in self._element.children:
            klass = getattr(child.__class__, 'RENDERED_ELEMENT_CLASS', RenderedElement)

            if child.xml_tag is None:  # type: AnonymusElement
                yield child.text
            elif for_loop_def := child.xml_attrs.get(f'{{{constants.XML_NAMESPACE_FLOW_CONTROL}}}for', None):  # type: ForLoop
                for counter, loop_var_name, loop_var_val in for_loop_def.eval(self._context):
                    self._context['loop'] = {'index': counter, 'index0': counter - 1, 'odd': bool(counter % 2 == 1)}
                    self._context[loop_var_name] = loop_var_val

                    if_def = child.xml_attrs.get(f'{{{constants.XML_NAMESPACE_FLOW_CONTROL}}}if', None)
                    if if_def is None or if_def.eval(self._context):
                        yield klass(element=child, context=self._context)
            else:
                if_def = child.xml_attrs.get(f'{{{constants.XML_NAMESPACE_FLOW_CONTROL}}}if', None)
                if if_def is None or if_def.eval(self._context):
                    yield klass(element=child, context=self._context)

    def to_json(self):
        retval = [self.node_name, self.node_attributes, []]
        for child in self.children:
            if isinstance(child, str):
                retval[2].append(child)
            elif hasattr(child, 'to_json'):
                retval[2].append(child.to_json())
            else:
                retval[2].append(str(child))

        return retval

    @contextmanager
    def frame(self, io, indent):
        indent = constants.INDENT * indent

        xml_attrs = ''
        for k, v in self.node_attributes.items():
            xml_attrs += f' {k}="{v}"'

        prefix = '<!DOCTYPE HTML>' if self.node_name == 'html' else ''
        io.write(f'{indent}{prefix}<{self.node_name}{xml_attrs}>\n')
        yield
        io.write(f'{indent}</{self.node_name}>\n')

    def to_xml(self, io=None, indent=0):
        io = io or StringIO()
        with self.frame(io, indent):
            for child in self.children:
                if isinstance(child, str):
                    io.write(f'{constants.INDENT * (indent + 1)}{child}\n')
                else:
                    if hasattr(child._element, 'to_xml'):
                        child._element.to_xml(io, indent + 1, child)
                    else:
                        child.to_xml(io, indent + 1)

        return io.getvalue()

    def to_jsx(self, io=None):
        io = io or StringIO()
        io.write(f"k('{self.node_name}', {{")
        for k, v in self.node_attributes.items():
            if k == 'class':
                key = 'className'
            elif k.startswith(f'{{{constants.XML_NAMESPACE_POJO}}}'):
                key = k.replace(f'{{{constants.XML_NAMESPACE_POJO}}}', '')
                io.write(f"'{key}': {v},")
                continue
            else:
                key = k

            if hasattr(v, '__iter__') and not isinstance(v, str):
                v = '[' + ','.join(f"'{_}'" for _ in v) + ']'
            elif v is None:
                v = "null"
            elif isinstance(v, bool):
                v = 'true' if v else 'false'
            else:
                v = json.dumps(v, ensure_ascii=False).replace('\\"', "'")

            io.write(f"'{key}': {v},")
        io.write('}, [')

        for child in self.children:
            if hasattr(child, 'to_jsx'):
                child.to_jsx(io)
            else:
                io.write(json.dumps(str(child), ensure_ascii=False).replace('\\"', "'"))

            io.write(',')

        io.write('])')
        return io.getvalue()

    def to_js_template(self, io=None):
        io = io or StringIO()
        io.write(f"<{self.node_name} ")
        for k, v in self.node_attributes.items():
            if k.startswith(f'{{{constants.XML_NAMESPACE_POJO}}}'):
                #key = k.replace(f'{{{constants.XML_NAMESPACE_POJO}}}', '')
                #io.write(f"'{key}': {v},")
                continue
            else:
                key = k

            if hasattr(v, '__iter__') and not isinstance(v, str):
                v = '[' + ','.join(f"'{_}'" for _ in v) + ']'
            elif v is None:
                v = "null"
            elif isinstance(v, bool):
                v = 'true' if v else 'false'
            else:
                v = json.dumps(v, ensure_ascii=False).replace('\\"', "'")

            if isinstance(v, (int, float, str)):
                if k.startswith('.'):  # property binding, string quote must be cutted
                    v = v[1:-1]
                io.write(f"{key}={v} ")

        if (self.node_name.startswith('kr-') and self.node_name != 'kr-app') or self.node_name.startswith('sm-')  or self.node_name.startswith('fx-'):  # TODO: ennél robosztusabb megoldást...
            io.write('.$children=${m.html`')
            for child in self.children:
                if hasattr(child, 'to_js_template'):
                    child.to_js_template(io)
                else:
                    io.write(str(child).strip().replace('\\', '\\\\'))
            io.write("`}")
            io.write('>')
        else:
            io.write('>')
            for child in self.children:
                if hasattr(child, 'to_js_template'):
                    child.to_js_template(io)
                else:
                    io.write(str(child).replace('\\', '\\\\'))

        if self.node_name != 'br':  # /br nem valid, TODO: kicsit robosztusabb ellenőrzés (pl.: has closing tag)
            io.write(f'</{self.node_name}>')
        io.flush()
        return io

    def to_psx(self):
        args = [ast.Constant(value=self.node_name), ast.Dict(keys=[], values=[]), ast.List(elts=[])]

        values = []
        for v in self.node_attributes.values():
            if isinstance(v, Bind):
                tree = ast.parse(v.expr)
                values.append(ast.Lambda(
                    args=ast.arguments(args=[ast.arg(arg='self')], posonlyargs=[], defaults=[], kwarg=None, kw_defaults=[], kwonlyargs=[], vararg=None),
                    body=tree.body[0].value))
            else:
                values.append(ast.Constant(v))

        args[1].keys = [ast.Constant(value=_.replace(f'{{{constants.XML_NAMESPACE_NOOP}}}', '')) for _ in self.node_attributes.keys()]
        args[1].values = values

        for child in self.children:
            if isinstance(child, str):
                args[2].elts.append(ast.Constant(value=child))
            else:
                args[2].elts.append(child.to_psx())

        node = ast.Call(
            func=ast.Name(id='h'),
            args=args,
            keywords=[]
        )

        return node

    def serialize(self, output='json', **kwargs):
        if output == 'json':
            return self.to_json()
        elif output == 'xml':
            return self.to_xml()
        elif output == 'jsx':
            return self.to_jsx()
        elif output == 'psx':
            return ast.unparse(self.to_psx())
        elif output == 'js_template':
            include_root = kwargs.get('include_root', True)
            if include_root is False:
                io = StringIO()
                for child in self.children:
                    child.to_js_template(io)
                return 'm.html`' + io.getvalue() + '`'
            else:
                return self.to_js_template(**kwargs).getvalue()
        else:
            raise NotImplementedError(f'Invalid output format: {output}')
